from mtsa import (
    MFCCMix,
    Hitachi,
    FEATURES,
    files_train_test_split,
    calculate_aucroc,
    get_tsne_results,
    files_train_test_split_combined
)
from sklearn.model_selection import LeaveOneOut
from sklearn.mixture import GaussianMixture
import seaborn as sns
import matplotlib.pyplot as plt
import glob
from functools import reduce
import itertools as ite
import numpy as np
import argparse
import os
import json
from scripts.common import elapsed_time, multiple_runs
import time
import functools as ft
from scripts.experiments import *


class PaperScript2023:

    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)

    def get_features(self):
        number_features = np.arange(1, len(FEATURES)+1)
        def combinations(r): return ite.combinations(FEATURES, r)
        all_combinations = map(combinations, number_features)
        features = reduce(lambda x, y: ite.chain(x, y), all_combinations)
        return features

    def get_models_mfccmix(self):
        def build_mfccmix_by_features(features):
            mfccmix = MFCCMix(
                features=list(features),
                final_model=GaussianMixture(n_components=self.n_components)
            )
            return ModelParam("MFCCMix " + "+".join([f[0] for f in features]), mfccmix)
        features = self.get_features()
        mfcc_models = map(build_mfccmix_by_features, features)
        return mfcc_models

    def get_models_baseline(self):
        hitachi = [ModelParam("Baseline", Hitachi())]
        return hitachi

    def get_models(self):
        model_baseline = self.get_models_baseline()
        models_mfccmix = self.get_models_mfccmix()
        all_models = reduce(lambda x, y: ite.chain(
            x, y), (model_baseline, models_mfccmix))
        return list(all_models)

    def get_paths(self, level=None):
        paths = glob.glob(os.path.join(self.path, f'*{os.sep}' * level))
        return paths
    
    def get_params(self, level=None):
        if not level:
            level = self.level
        all_models = self.get_models()
        all_paths = self.get_paths(level)
        all_params = list(ite.product(all_paths, all_models))
        all_params = [Params(*p) for p in all_params]
        return all_params
                
    def write_json(self, result : ParamResult, **kwargs):
        path_output = os.path.join(self.output, 'rocs')
        if not os.path.exists(path_output):
            os.makedirs(path_output)
        filename = 'filename'
        if 'filename' in kwargs:
            filename = kwargs['filename']
        filepath = os.path.join(path_output, f'{filename}')
        
        ts = time.time_ns()
        filepath = f'{filepath}_{ts}'
        with open(filepath, "a") as jf:
            #TODO this is ugly but leting keras to be serialized causes circular reference at json.dumps 
            should_dump = lambda o: any(map(ft.partial(isinstance, o), [t for t in [int, float, str, ParamResult, Result, Params, ModelParam]]))
            def default(o):
                if should_dump(o):
                    if hasattr(o, "__dict__"):
                        return o.__dict__
                    else:
                        return str(o)
                else:
                    return ""
            json_data = json.dumps(result, indent=4, default=default)
            print(json_data, file=jf)

    def fit_individual_model(self, params: Params):
        # print(params)
        path = params.path
        model = params.model.obj
        model_name = params.model.name
        X_train, X_test, y_train, y_test = files_train_test_split(path)
        result_fit = elapsed_time(model.fit, X_train, y_train)
        param_result_fit = ParamResult(params, result_fit)
        self.write_json(param_result_fit, filename='results_individual_fit')
        result_roc = elapsed_time(calculate_aucroc, model, X_test, y_test)
        param_result_roc = ParamResult(params,result_roc)
        self.write_json(param_result_roc, filename='results_individual_roc')
        x, y, hue = self.local_get_tsne_results(model, X_test, y_test)
        self.plot_tsne(model_name, path, x, y, hue)

    def local_get_tsne_results(self, model, X_test, y_test):
        x, y, hue = get_tsne_results(
            model=model,
            X=model.transform(X_test),
            y=y_test,
            perplexity=self.perplexity
        )
        return  x, y, hue
    
    def plot_tsne(self, model_name, path, x, y, hue):
        plt.figure(figsize=(16, 10))
        sns.scatterplot(
            x=x, y=y,
            hue=hue,
            palette=sns.color_palette("hls", 2),
            legend="full",
            s=100
            # alpha=0.3
        )
        path_output = os.path.join(self.output, 'results_tse')
        if not os.path.exists(path_output):
            os.makedirs(path_output)
        filename = '_'.join(path.split(os.path.sep)[:-1][-(self.level+1):]) + "_" + model_name
        ts = time.time_ns()
        filename = f'{filename}_{ts}'
        plt.savefig(os.path.join(path_output, f'{filename}.pdf'))

   

    def results_individual(self):
        print("START: results_individual")
        all_params = self.get_params(self.level)
        runs = self.runs
        results = [multiple_runs(runs, self.fit_individual_model, param) for param in all_params]        
        print("FINISH: results_individual")



    def calculate_roc(self, params_tuple):
        paths, model_name, model, split = params_tuple
        path_train = paths[split[0]]
        path_test = paths[split[1]]
        print(
            f"results_combined calculate_roc {str(path_test)} {model_name}")
        train_X_train, train_X_test, train_y_train, train_y_test = files_train_test_split_combined(
            path_train)
        test_X_train, test_X_test, test_y_train, test_y_test = files_train_test_split_combined(
            path_test)
        params = Params(path_test[0], ModelParam(model_name, model))
        result_fit = elapsed_time(model.fit, train_X_train, train_y_train)
        param_result_fit = ParamResult(params, result_fit)
        self.write_json(param_result_fit, filename='results_combined_fit')
        result_roc = elapsed_time(calculate_aucroc, model, test_X_test, test_y_test)
        param_result_roc = ParamResult(params,result_roc)
        self.write_json(param_result_roc, filename='results_combined_roc')
        x, y, hue = self.local_get_tsne_results(model, test_X_test, test_y_test)
        self.plot_tsne(model_name, params.path, x, y, hue)

    def calculate_roc_combined(self, params: Params):
        path = params.path
        model = params.model.obj
        model_name = params.model.name
        # print(
        #     f"results_combined calculate_roc_combined {path} {model_name}")
        paths = np.array(glob.glob(os.path.join(path, f'*{os.sep}')))
        loo = LeaveOneOut()
        splits = loo.split(paths)
        rocs = map(self.calculate_roc, [
                (paths, model_name, model, split) for split in splits])
        return list(rocs)

    # def results_combined(self):
    #     print("START: results_combined")
        
    #     all_params = self.get_params(self.level -1 )

    #     results = [
    #         multiple_runs(self.runs, 
    #         self.calculate_roc_combined, 
    #         param) 
    #         for param in all_params
    #         ]   
        
    #     print("FINISH: results_combined")


# def main(parse_args):
#     paper_script_2023 = PaperScript2023(**parse_args.__dict__)
#     # paper_script_2023.results()


# if __name__ == "__main__":
#     print("MTSA Pipeline")
#     print("Initializing...")
#     parser = argparse.ArgumentParser()
#     parser.add_argument('--type', type=str)
#     parser.add_argument('--sampling_rate', type=int)
#     parser.add_argument('--path', type=str)
#     parser.add_argument('--n_components', type=int)
#     parser.add_argument('--level', type=int)
#     parser.add_argument('--perplexity', type=int)
#     parser.add_argument('--runs', type=int)
#     parser.add_argument('--output', type=str)
#     parse_args = parser.parse_args()
#     main(parse_args)
#     print("MTSA Pipeline: Finished")

# def convert(match_obj):
#             if match_obj.group(0) == '_':
#                 return ""
#             if match_obj.group(0) == '/':
#                 return " "



import numpy as np
import abc

class CorrelationNetwork(abc.ABC):
    """
    Base class for implementing correlation between mfccs
    """

    def get_correlation(self, X: np.ndarray):
        """Implement correlation"""

class PearsonCorrelationNetwork(CorrelationNetwork):

    def get_correlation(self, X : np.ndarray):
        C = np.corrcoef(X, rowvar=True)
        return C

class DynamicCorrelationNetwork(CorrelationNetwork):

    def get_correlation(self, X : np.ndarray):
        """
        Calculate dynamic correlation <(i - <i>)(j - <j>)>
        where <x> is the first moment (e.g, expected value)
        """
        X = X.T
        x_min = np.min(X, axis=0)
        x_max = np.max(X, axis=0)
        X = (X - x_min) / (x_max - x_min)
        X = 2*X - 1
        m = np.mean(X, axis=0)
        D = X - m
        return 1/len(D) * D.T.dot(D)
    
import scipy.stats as st
class SpearmanCorrelationNetwork(CorrelationNetwork):

    def get_correlation(self, X : np.ndarray):
        rho, pval = st.spearmanr(X)
        return rho


import time
from functools import partial
from functools import reduce
from mtsa.experiments import *

def elapsed_time(fun, *args) -> Result:
    start = time.time()
    fun_return = fun(*args)
    end = time.time()
    time_elapsed = end - start
    result = Result(time_elapsed, fun_return) 
    return result


def multiple_runs(runs, fun, *args):
    list_run_ids = range(runs)
    multiple_results = map(
        lambda run_id:
        {
            'run_id': run_id,
            "run_details": elapsed_time(fun, *args)
        },
        list_run_ids
        )

    return list(multiple_results)


import json
import functools as ft
def print(element):
    #TODO this is ugly but leting keras to be serialized causes circular reference at json.dumps 
    should_dump = lambda o: any(map(ft.partial(isinstance, o), [t for t in [int, float, str, ParamResult, Result, Params, ModelParam]]))
    def default(o):
        if should_dump(o):
            if hasattr(o, "__dict__"):
                return o.__dict__
            else:
                return str(o)
        else:
            return ""
    json_data = json.dumps(element, indent=4, default=default)
    return json


import torch.nn as nn
import torch.nn.functional as F
import torch

class GNN(nn.Module):
    """
    The GNN module applied in GANF
    """
    def __init__(self, input_size, hidden_size):
        super(GNN, self).__init__()
        self.lin_n = nn.Linear(input_size, hidden_size)
        self.lin_r = nn.Linear(input_size, hidden_size, bias=False)
        self.lin_2 = nn.Linear(hidden_size, hidden_size)

    def forward(self, h, adjacent_matrix):
        ## adjacent_matrix: K X K
        ## H: N X K  X L X D

        h_n = self.lin_n(torch.einsum('nkld,kj->njld',h,adjacent_matrix))
        h_r = self.lin_r(h[:,:,:-1])
        h_n[:,:,1:] += h_r
        h = self.lin_2(F.relu(h_n))

        return h


import torch.nn as nn
import torch
import random
import numpy as np
import pandas as pd
from torch.nn.init import xavier_uniform_
from torch.nn.utils import clip_grad_value_
from torch.utils.data import DataLoader
from mtsa.models.GANF_components.NF import MAF, RealNVP
from mtsa.models.GANF_components.ganfLayoutData import GANFData
from mtsa.models.GANF_components.gnn import GNN

class GANFBaseModel(nn.Module):
    def __init__ (self, 
                  n_blocks = 6,
                  input_size = 1,
                  hidden_size = 32,
                  n_hidden = 1,
                  dropout = 0.0,
                  model="MAF",
                  batch_norm=False,
                  rho_max = float(1e16),
                  max_iteraction = 2,
                  learning_rate = float(1e-3),
                  alpha = 0.0,
                  weight_decay = float(5e-4),
                  epochs = 20,
                  device = None,
                  index_CUDA_device = '0'
                  ):
        super().__init__()
        self.min = 0
        self.max = 0 
        self.adjacent_matrix = None
        self.rho_max = rho_max     
        self.max_iteraction = max_iteraction    
        self.learning_rate = learning_rate          
        self.alpha = alpha       
        self.weight_decay= weight_decay
        self.epochs = epochs
        self.hidden_state = None

        if device != None: 
            self.device = device
        else:
            self.device = torch.device(f'cuda:{str(index_CUDA_device)}' if torch.cuda.is_available() else "cpu")

        #Reducing dimensionality 
        self.rnn = nn.LSTM(input_size=input_size,hidden_size=hidden_size,batch_first=True, dropout=dropout, device=self.device)

        # Graph Neural Networks model
        self.gcn = GNN(input_size=hidden_size, hidden_size=hidden_size) 

        #probability density estimation
        if model=='MAF':
            #Masked auto regressive flow
            self.nf = MAF(n_blocks, input_size, hidden_size, n_hidden, cond_label_size=hidden_size, batch_norm=batch_norm,activation='tanh')
        else:
            #real-valued non-volume preserving (real NVP)
            self.nf = RealNVP(n_blocks, input_size, hidden_size, n_hidden, cond_label_size=hidden_size, batch_norm=batch_norm)

        self.to(device=self.device)

    @property
    def name(self):
        return "GANFBaseModel " + "+".join([f[0] for f in self.features])
      
    def get_adjacent_matrix(self):
        return self.adjacent_matrix

    def fit(self, X, y=None, batch_size = None, epochs= None, max_iteraction= None, learning_rate = None, debug_dataframe= None, isWaveData = False, mono = None):
        torch.cuda.manual_seed(10)
        random.seed(10)
        np.random.seed(10)
        torch.manual_seed(10)
        torch.autograd.set_detect_anomaly(True)

        self.isMonoData = mono
        loss_best = 100
        self.isWaveData = isWaveData
        h_A_old = np.inf
        h_tol = float(1e-6)
        rho = 1.0

        if epochs is not None:
           self.epochs = epochs
        if max_iteraction is not None:
            self.max_iteraction =  max_iteraction
        if learning_rate is not None:
            self.learning_rate = learning_rate
        if batch_size is None:
            batch_size = 32

        dimension = X.data.shape[1]
        self.adjacent_matrix = self.__init_adjacent_matrix(X.data)

        dataloaders = self.__create_dataLoader(torch.tensor(X), batch_size=batch_size, isMonoData = mono)
        adjacent_matrix = self.adjacent_matrix

        self.__fitCore(loss_best, h_A_old, h_tol, dimension, dataloaders, adjacent_matrix, rho)

    def predict(self, X):
        return self.__forward(X, self.adjacent_matrix)

    def score_samples(self, X):
        if self.isWaveData:
            return self.__score_wave_data(X=X)
        return self.__score_discrete_data(X=X)
    
    def __forward(self, x, A):
        # x: N X K X L X D 
        full_shape = x.shape

        # reshape: N*K, L, D
        x = x.reshape((x.shape[0]*x.shape[1], x.shape[2], x.shape[3]))
        h,_ = self.rnn(x)

        # resahpe: N, K, L, H
        h = h.reshape((full_shape[0], full_shape[1], h.shape[1], h.shape[2]))

        h = self.gcn(h, A)

        # reshappe N*K*L,H
        h = h.reshape((-1,h.shape[3]))
        x = x.reshape((-1,full_shape[3]))

        log_prob = self.nf.log_prob(x,h).reshape([full_shape[0],-1])#*full_shape[1]*full_shape[2]
        log_prob = log_prob.mean(dim=1)
        log_prob_x = log_prob.mean()
        return log_prob_x

    def __fitCore(self, loss_best, h_A_old, h_tol, dimension, dataloaders, adjacent_matrix, rho):
        previous_total_loss = None
        for j in range(self.max_iteraction):
            print('iteraction ' + str(j+1) + ' of ' + str(self.max_iteraction))

            while rho < self.rho_max:
                learning_rate = self.learning_rate #np.math.pow(0.1, epoch // 100)    
                optimizer = torch.optim.Adam([
                    {'params': self.parameters(), 'weight_decay':self.weight_decay},
                    {'params': [adjacent_matrix]}], lr=learning_rate, weight_decay=0.0)

                for epoch in range(self.epochs):
                    loss_train = []
                    epoch += 1
                    self.train()

                    for dataloader in dataloaders:
                        for x in dataloader:
                            x = x.to(self.device)
                            
                            optimizer.zero_grad()
                            loss = -self.__forward(x, adjacent_matrix)
                            h = torch.trace(torch.matrix_exp(adjacent_matrix*adjacent_matrix)) - dimension
                            total_loss = loss + 0.5 * rho * h * h + self.alpha * h

                            h.to(self.device)
                            
                            total_loss.backward()
                            clip_grad_value_(self.parameters(), 1)
                            optimizer.step()
                            loss_train.append(loss.item())
                            adjacent_matrix.data.copy_(torch.clamp(adjacent_matrix.data, min=0, max=1))
                            
                            if np.mean(loss_train) < loss_best:
                                loss_best = np.mean(loss_train)
                                self.adjacent_matrix = adjacent_matrix

                            previous_total_loss = total_loss

                    print('Epoch: {}, train -log_prob: {:.2f}, h: {}'.format(epoch, np.mean(loss_train), h.item()))
                    
                del optimizer
                torch.cuda.empty_cache()

                if h.item() > 0.5 * h_A_old:
                    rho *= 10
                else:
                    break

            h_A_old = h.item()
            self.alpha += rho*h.item()

            if h_A_old <= h_tol or rho >= self.rho_max:
                break
    
    def __score_discrete_data(self, X):
        X = np.array(X)
        X = X.reshape(X.shape[0], X.shape[1], 1, 1)
        X = torch.tensor(X).float()
        X = X.to(self.device)
        result = self.predict(X=X)
        return float(result)
    
    def __score_wave_data(self, X):
        dataloaders = self.__create_dataLoader(X, window_size=1, isMonoData = self.isMonoData)
        result = []
        for dataloader in dataloaders:
            for x in dataloader:
                x = x.to(self.device)
                result.append(self.predict(X=x))
        return torch.tensor(result).mean()
       
    def __create_dataLoader(self, X, batch_size = 32, window_size = 12, isMonoData = True):
        X = self.__create_dataframe(X, isMonoData)
        dataloaders = []
        for x in X:
            x_dataloader = DataLoader(GANFData(x, window_size=window_size), batch_size=batch_size, shuffle=False, num_workers=0, persistent_workers=False)
            dataloaders.append(x_dataloader)
        return dataloaders
    
    def __create_dataframe(self, X, isMonoData):
        if self.isWaveData and not isMonoData:
            dataframes = []
            for x in X:
                x = x.T
                x_df = self.__array2Dataframe(x) 
                dataframes.append(x_df)
            return dataframes
        
        if self.isWaveData and isMonoData:
            X = X.T

        X_df =  self.__array2Dataframe(X)

        return [X_df]

    def __array2Dataframe(self, x):
        x_df = pd.DataFrame(x)
        x_df= x_df.reset_index()
        x_df = x_df.rename(columns={"index":"channel"})
        x_df["channel"] = pd.to_datetime(x_df["channel"], unit="s")
        x_df = x_df.set_index("channel")
        x_df = self.__normalize(x_df)
        return x_df

    def __normalize(self, X):
        mean = X.values.mean()
        std = X.values.std()
        X = (X - mean)/std
        return X

    def __init_adjacent_matrix(self, X):
        torch.cuda.manual_seed(10)
        random.seed(10)
        np.random.seed(10)
        torch.manual_seed(10)
        init = torch.zeros([X.shape[1], X.shape[1]])
        init = xavier_uniform_(init).abs()
        init = init.fill_diagonal_(0.0)
        return torch.tensor(init, requires_grad=True, device=self.device)



import os
import torch
import torch.nn as nn
import torch.nn.functional as F
from timeit import default_timer as timer
def ConvEncoder(activation = nn.LeakyReLU, in_channels:int = 3, n_c:int = 64,
                    k_size:int = 5):
    
    enc = nn.Sequential(*(nn.Conv1d(in_channels, n_c, k_size, stride=2, padding=2),
                                            nn.BatchNorm1d(n_c),
                                            activation(),
                                            nn.Conv1d(n_c, n_c*2, k_size, stride=2, padding=2),
                                            nn.BatchNorm1d(n_c*2),
                                            activation(),
                                            nn.Conv1d(n_c*2, n_c*4, k_size, stride=2, padding=2),
                                            nn.BatchNorm1d(n_c*4),
                                            activation()))
    return enc

def ConvDecoder(activation = nn.LeakyReLU, in_channels:int = 3, n_c:int = 64,
                    k_size:int = 5):

    decoder = nn.Sequential(*(nn.ConvTranspose1d(n_c*4, n_c*2, k_size, stride=2, padding=2, output_padding=0),
                                            torch.nn.BatchNorm1d(n_c*2),
                                            activation(),
                                            torch.nn.ConvTranspose1d(n_c*2, n_c, k_size,stride=2, padding=2, output_padding=1),
                                            torch.nn.BatchNorm1d(n_c),
                                            activation(),
                                            torch.nn.ConvTranspose1d(n_c, in_channels, k_size,stride=2, padding=2, output_padding=1)))
    return decoder

class CNNAE(torch.nn.Module):
    """Recurrent autoencoder"""

    def __init__(self,in_channels:int = 3, n_channels:int = 16,
                    kernel_size:int = 5):
        super(CNNAE, self).__init__()

        # Encoder and decoder argsuration
        activation = torch.nn.LeakyReLU
        self.in_channels = in_channels
        self.n_c = n_channels
        self.k_size = kernel_size

        self.encoder = ConvEncoder(activation, in_channels, n_channels, kernel_size)

        self.decoder = ConvDecoder(activation, in_channels, n_channels, kernel_size)


    def forward(self, x:torch.Tensor):

        z = self.encoder.forward(x)

        x_out = self.decoder.forward(z)

        return x_out

class R_Net(torch.nn.Module):

    def __init__(self, activation = torch.nn.LeakyReLU, in_channels:int = 3, n_channels:int = 16,
                    kernel_size:int = 5, std:float = 0.2):

        super(R_Net, self).__init__()

        self.activation = activation
        self.in_channels = in_channels
        self.n_c = n_channels
        self.k_size = kernel_size
        self.std = std

        self.Encoder = ConvEncoder(activation, in_channels, n_channels, kernel_size)

        self.Decoder = ConvDecoder(activation, in_channels, n_channels, kernel_size)

    def forward(self, x:torch.Tensor, noise:bool = True):

        x_hat = self.add_noise(x) if noise else x
        z = self.Encoder.forward(x_hat)

        x_out = self.Decoder.forward(z)

        return x_out

    def add_noise(self, x):

        noise = torch.randn_like(x) * self.std
        x_hat = x + noise

        return x_hat

class D_Net(torch.nn.Module):

    def __init__(self, in_resolution:int, activation = torch.nn.LeakyReLU, in_channels:int = 3, n_channels:int = 16, kernel_size:int = 5):

        super(D_Net, self).__init__()

        self.activation = activation
        self.in_resolution = in_resolution
        self.in_channels = in_channels
        self.n_c = n_channels
        self.k_size = kernel_size

        self.cnn = ConvEncoder(activation, in_channels, n_channels, kernel_size)

        # Compute output dimension after conv part of D network

        self.out_dim = self._compute_out_dim()

        self.fc = torch.nn.Linear(self.out_dim, 1)

    def _compute_out_dim(self):
        
        test_x = torch.Tensor(1, self.in_channels, self.in_resolution)
        for p in self.cnn.parameters():
            p.requires_grad = False
        test_x = self.cnn(test_x)
        out_dim = torch.prod(torch.tensor(test_x.shape[1:])).item()
        for p in self.cnn.parameters():
            p.requires_grad = True

        return out_dim

    def forward(self, x:torch.Tensor):

        x = self.cnn(x)

        x = torch.flatten(x, start_dim = 1)

        out = self.fc(x)

        return out

def R_Loss(d_net: torch.nn.Module, x_real: torch.Tensor, x_fake: torch.Tensor, lambd: float) -> dict:

    pred = d_net(x_fake)
    y = torch.ones_like(pred)

    rec_loss = F.mse_loss(x_fake, x_real)
    gen_loss = F.binary_cross_entropy_with_logits(pred, y) # generator loss

    L_r = gen_loss + lambd * rec_loss

    return {'rec_loss' : rec_loss, 'gen_loss' : gen_loss, 'L_r' : L_r}

def D_Loss(d_net: torch.nn.Module, x_real: torch.Tensor, x_fake: torch.Tensor) -> torch.Tensor:

    pred_real = d_net(x_real)
    pred_fake = d_net(x_fake.detach())
    
    y_real = torch.ones_like(pred_real)
    y_fake = torch.zeros_like(pred_fake)

    real_loss = F.binary_cross_entropy_with_logits(pred_real, y_real)
    fake_loss = F.binary_cross_entropy_with_logits(pred_fake, y_fake)

    return real_loss + fake_loss

# Wasserstein GAN loss (https://arxiv.org/abs/1701.07875)

def R_WLoss(d_net: torch.nn.Module, x_real: torch.Tensor, x_fake: torch.Tensor, lambd: float) -> dict:

    pred = torch.sigmoid(d_net(x_fake))

    rec_loss = F.mse_loss(x_fake, x_real)
    gen_loss = -torch.mean(pred) # Wasserstein G loss: - E[ D(G(x)) ]

    L_r = gen_loss + lambd * rec_loss

    return {'rec_loss' : rec_loss, 'gen_loss' : gen_loss, 'L_r' : L_r}

def D_WLoss(d_net: torch.nn.Module, x_real: torch.Tensor, x_fake: torch.Tensor) -> torch.Tensor:

    pred_real = torch.sigmoid(d_net(x_real))
    pred_fake = torch.sigmoid(d_net(x_fake.detach()))
    
    dis_loss = -torch.mean(pred_real) + torch.mean(pred_fake) # Wasserstein D loss: -E[D(x_real)] + E[D(x_fake)]

    return dis_loss

# %%
def train_model(r_net: torch.nn.Module,
                d_net: torch.nn.Module,
                train_dataset: torch.utils.data.Dataset,
                valid_dataset: torch.utils.data.Dataset,
                r_loss = R_Loss,
                d_loss = D_Loss,
                lr_scheduler = None,
                optimizer_class = torch.optim.Adam,
                optim_r_params: dict = {},
                optim_d_params: dict = {},
                learning_rate: float = 0.001,
                scheduler_r_params: dict = {},
                scheduler_d_params: dict = {},
                batch_size: int = 1024,
                max_epochs: int = 40,
                epoch_step: int = 1,
                save_step: int = 5,
                lambd: float = 0.2,
                device: torch.device = torch.device('cuda'),
                save_path: str = ".") -> tuple:

    optim_r = optimizer_class(r_net.parameters(), lr = learning_rate, **optim_r_params)
    optim_d = optimizer_class(d_net.parameters(), lr = learning_rate, **optim_d_params)

    if lr_scheduler:
        scheduler_r = lr_scheduler(optim_r, **scheduler_r_params)
        scheduler_d = lr_scheduler(optim_d, **scheduler_d_params)

    train_loader = torch.utils.data.DataLoader(train_dataset, shuffle=True, batch_size=batch_size)
    valid_loader = torch.utils.data.DataLoader(valid_dataset, batch_size=batch_size)

    for epoch in range(max_epochs):

        start = timer()
        train_metrics = train_single_epoch(r_net, d_net, optim_r, optim_d, r_loss, d_loss, train_loader, lambd, device)
        valid_metrics = validate_single_epoch(r_net, d_net, r_loss, d_loss, valid_loader, device)
        time = timer() - start


        if epoch % epoch_step == 0:
            print(f'Epoch {epoch}:')
            print('Train Metrics:', train_metrics)
            print('Val Metrics:', valid_metrics)
            print(f'TIME: {time:.2f} s')

        if lr_scheduler:
            scheduler_r.step()
            scheduler_d.step()

        if epoch % save_step == 0:
            torch.save(r_net.state_dict(), os.path.join(save_path, "r_net_{}.pt".format(epoch)))
            torch.save(d_net.state_dict(), os.path.join(save_path, "d_net_{}.pt".format(epoch)))
            print(f'Saving model on epoch {epoch}')

    return (r_net, d_net)

def train_single_epoch(r_net, d_net, optim_r, optim_d, r_loss, d_loss, train_loader, lambd, device) -> dict:

    r_net.train()
    d_net.train()

    train_metrics = {'rec_loss' : 0, 'gen_loss' : 0, 'dis_loss' : 0}

    for data in train_loader:

        x = data.to(device)
        x = torch.transpose(x, dim0=2, dim1=3)
        x_real = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])
        
        x_fake = r_net(x_real)
        d_net.zero_grad()

        dis_loss = d_loss(d_net, x_real, x_fake)

        dis_loss.backward()
        optim_d.step()

        r_net.zero_grad()

        r_metrics = r_loss(d_net, x_real, x_fake, lambd) # L_r = gen_loss + lambda * rec_loss

        r_metrics['L_r'].backward()
        optim_r.step()

        train_metrics['rec_loss'] += r_metrics['rec_loss']
        train_metrics['gen_loss'] += r_metrics['gen_loss']
        train_metrics['dis_loss'] += dis_loss

    train_metrics['rec_loss'] = train_metrics['rec_loss'].item() / (len(train_loader.dataset) / train_loader.batch_size)
    train_metrics['gen_loss'] = train_metrics['gen_loss'].item() / (len(train_loader.dataset) / train_loader.batch_size)
    train_metrics['dis_loss'] = train_metrics['dis_loss'].item() / (len(train_loader.dataset) / train_loader.batch_size)

    return train_metrics

def validate_single_epoch(r_net, d_net, r_loss, d_loss, valid_loader, device) -> dict:

    r_net.eval()
    d_net.eval()

    valid_metrics = {'rec_loss' : 0, 'gen_loss' : 0, 'dis_loss' : 0}

    with torch.no_grad():
        for data in valid_loader:

            x = data.to(device)
            x = torch.transpose(x, dim0=2, dim1=3)
            x_real = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])

            x_fake = r_net(x_real)

            dis_loss = d_loss(d_net, x_real, x_fake)

            r_metrics = r_loss(d_net, x_real, x_fake, 0)
                
            valid_metrics['rec_loss'] += r_metrics['rec_loss']
            valid_metrics['gen_loss'] += r_metrics['gen_loss']
            valid_metrics['dis_loss'] += dis_loss

        valid_metrics['rec_loss'] = valid_metrics['rec_loss'].item() / (len(valid_loader.dataset) / valid_loader.batch_size)
        valid_metrics['gen_loss'] = valid_metrics['gen_loss'].item() / (len(valid_loader.dataset) / valid_loader.batch_size)
        valid_metrics['dis_loss'] = valid_metrics['dis_loss'].item() / (len(valid_loader.dataset) / valid_loader.batch_size)

        return valid_metrics



# %%



from mtsa import MFCCMix

from sklearn.model_selection import (
    RandomizedSearchCV,
    ShuffleSplit
)

from sklearn.metrics import (
    make_scorer,
    recall_score
)

import numpy as np


class MFCCMixCV(MFCCMix):
    
    def __init__(
        self,
        sampling_rate = 16000,
        random_state = None,
        cv=None) -> None:
        super(MFCCMixCV, self).__init__(
            random_state=random_state,
            sampling_rate=sampling_rate
            )
        if not cv:
            cv = ShuffleSplit(
                random_state=random_state,
                n_splits=5)
            self.cv = cv
        
        params = [
                {
                    "final_model__covariance_type": ['full', 'tied', 'diag', 'spherical'],
                    # "final_model__n_components": [n for n in np.arange(1, 230)]
                    # "final_model__covariance_type": ['full'],
                    "final_model__n_components": np.arange(230, 231)
                },
        ]
        
        self.params = params

        def my_score(y_true, y_pred):
            y_pred = 1*(y_pred == 1)
            return recall_score(y_true=y_true, y_pred=y_pred)
        my_scorer = make_scorer(my_score)
        search = RandomizedSearchCV(
            n_iter=4,
            estimator=self.model,
            param_distributions=self.params,
            scoring=my_scorer,
            # scoring=my_score,
            cv=self.cv
        )
        self.search = search

    def fit(self, X, y=None, **fit_params):
        return self.search.fit(X, y)

    def transform(self, X, y=None, **fit_params):
        return self.search.transform(X, y, **fit_params)

    def predict(self, X):
        return self.search.predict(X)

    def score_samples(self, X):
        return self.search.score_samples(X=X)


"""Base classes for all mfcc estimators."""

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from functools import reduce
import itertools as ite

class MagnitudeMeanFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = X.mean(axis=2)
        return Xt
    
class MagnitudeStdFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = X.std(axis=2)
        return Xt
    
from mtsa.correlation_networks import PearsonCorrelationNetwork
class CorrelationFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        def get_triu(G):
            return G[np.triu_indices(len(G), k = 1)]
        corr = PearsonCorrelationNetwork()
        Xt = map(corr.get_correlation, X)
        Xt = map(get_triu, Xt)
        Xt= np.array(list(Xt)) 
        return Xt
    

FEATURES = [
    ("M", MagnitudeMeanFeatureMfcc()), 
    ("S", MagnitudeStdFeatureMfcc()), 
    ("C", CorrelationFeatureMfcc())
]

def get_features(self):
    number_features = np.arange(1, len(FEATURES)+1)
    def combinations(r): return ite.combinations(FEATURES, r)
    all_combinations = map(combinations, number_features)
    features = reduce(lambda x, y: ite.chain(x, y), all_combinations)
    return features




from mtsa import (
    HitachiDCASE2020,
    
)
from mtsa.utils import files_train_test_split_dcase2020_task2

from mtsa.metrics import (
    calculate_aucroc
)

path = '/data/DCASE2020/dev/fan'
X_train, X_test, y_train, y_test = files_train_test_split_dcase2020_task2(path, pattern='id_0[0|4|6]')

model = HitachiDCASE2020()

model.fit(X_train, y_train)

roc = calculate_aucroc(model, X_test, y_test)

print(roc)


import numpy as np
from mtsa.features.mel import Array2Mfcc
from mtsa.utils import Wav2Array
from typing import List, Optional
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, OutlierMixin
from mtsa.models.ransyncoders_components.ransyncoders_base import RANSynCodersBase

class RANSynCoders(BaseEstimator, OutlierMixin):
    def __init__(self, 
                 n_estimators: int = 5,
                 max_features: int = 5,
                 encoding_depth: int = 1,
                 latent_dim: int = 1,
                 decoding_depth: int = 2,
                 activation: str = 'relu',
                 output_activation: str = 'sigmoid',
                 delta: float = 0.05,  # quantile bound for regression
                 # Syncrhonization inputs
                 synchronize: bool = False, #False
                 force_synchronization: bool = True,  # if synchronization is true but no significant frequencies found
                 min_periods: int = 3,  # if synchronize and forced, this is the minimum bound on cycles to look for in train set
                 freq_init: Optional[List[float]] = None,  # initial guess for the dominant angular frequency
                 max_freqs: int = 5,  # the number of sinusoidal signals to fit
                 min_dist: int = 60,  # maximum distance for finding local maximums in the PSD
                 trainable_freq: bool = False,  # whether to make the frequency a variable during layer weight training
                 bias: bool = True,  # add intercept (vertical displacement)
                 sampling_rate: int = 16000,
                 mono: bool = False,
                 is_acoustic_data: bool = False,
                 normal_classifier: int = 0,
                 abnormal_classifier: int = 1,
                ) -> None:
            super().__init__()
            # Rancoders inputs:
            self.n_estimators = n_estimators
            self.max_features = max_features
            self.encoding_depth = encoding_depth
            self.latent_dim = latent_dim
            self.decoding_depth = decoding_depth
            self.activation = activation
            self.output_activation = output_activation
            self.delta = delta
        
            # Syncrhonization inputs
            self.synchronize = synchronize
            self.force_synchronization = force_synchronization
            self.min_periods = min_periods
            self.freq_init = freq_init  # in radians (angular frequency)
            self.max_freqs = max_freqs
            self.min_dist = min_dist
            self.trainable_freq = trainable_freq
            self.bias = bias
            self.sampling_rate = sampling_rate
            self.mono = mono
            self.is_acoustic_data = is_acoustic_data
            self.normal_classifier = normal_classifier
            self.abnormal_classifier = abnormal_classifier
            self.model = self.build_model()

    @property
    def name(self):
        return "RANSynCoder " + "+".join([f[0] for f in self.features])
        
    def fit(self, 
            X,
            y = None, 
            timestamps_matrix: np.ndarray = None,
            batch_size: int = 180, 
            learning_rate: float = 0.001,
            epochs: int = 10,
            freq_warmup: int = 5,  # number of warmup epochs to prefit the frequency
            sin_warmup: int = 5,  # number of warmup epochs to prefit the sinusoidal representation 
            pos_amp: bool = True,  # whether to constraint amplitudes to be +ve only
            shuffle: bool = True
        ):
        return self.model.fit(X, 
                              y, 
                              final_model__timestamps_matrix = timestamps_matrix,
                              final_model__batch_size = batch_size, 
                              final_model__learning_rate = learning_rate,
                              final_model__epochs = epochs,
                              final_model__freq_warmup = freq_warmup,
                              final_model__sin_warmup = sin_warmup,
                              final_model__pos_amp = pos_amp,
                              final_model__shuffle = shuffle,
                            )
    
    def predict(self, X):
        return self.model.predict(X)

    def score_samples(self, X):
        if self.is_acoustic_data:    
            return np.array(
                list(
                    map(
                        self.model.score_samples, 
                        [[x] for x in X])
                    )
                )
            
        return self.model.score_samples(X)
    
    def set_timestamps_matrix_to_predict(self, timestamps_matrix: np.ndarray):
        self.final_model.set_timestamps_matrix_to_predict(timestamps_matrix)
    
    def save(self):
        return self.final_model.save()

    def get_config(self):
        return self.final_model.get_config()

    def get_final_model(self):
        return RANSynCodersBase(n_estimators = self.n_estimators, 
                                max_features = self.max_features, 
                                encoding_depth = self.encoding_depth, 
                                latent_dim = self.latent_dim, 
                                decoding_depth = self.decoding_depth, 
                                activation = self.activation,
                                output_activation = self.output_activation,
                                delta = self.delta,
                                synchronize = self.synchronize,
                                force_synchronization = self.force_synchronization,
                                min_periods = self.min_periods,  
                                freq_init = self.freq_init,  
                                max_freqs = self.max_freqs,  
                                min_dist = self.min_dist, 
                                trainable_freq = self.trainable_freq,  
                                bias = self.bias,
                                sampling_rate = self.sampling_rate,
                                mono = self.mono,
                                is_acoustic_data = self.is_acoustic_data,
                                normal_classifier = self.normal_classifier,
                                abnormal_classifier = self.abnormal_classifier
                            )
        
    def build_model(self):
        steps = self.__get_pipeline_steps()
        model = Pipeline(steps=steps)
        
        return model
    
    def __get_pipeline_steps(self):
        wav2array = Wav2Array(sampling_rate=self.sampling_rate, mono=self.mono)
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
       
        self.final_model = self.get_final_model()
        
        if self.is_acoustic_data:
            steps = [("wav2array", wav2array),
                     ("array2mfcc", array2mfcc),
                     ("final_model", self.final_model)
                    ]
        else:
            steps = [("final_model", self.final_model)]
        
        return steps
    



from mtsa import Hitachi

from mtsa.utils import (
    files_train_test_split
)
from mtsa.metrics import (
    calculate_aucroc
)
                         
model = Hitachi()

path = "/data/MIMII/fan/id_00"
X_train, X_test, y_train, y_test = files_train_test_split(path)

model.fit(X_train, y_train)

roc = calculate_aucroc(model, X_test, y_test)

print(roc)